/*----------------------------------------------------------------------
  File    : whamster.c
  Contents: main program for hamster (Windows 95 version)
  Author  : Christian Borgelt
  History : 14.10.1997 file created
            16.10.1997 first version completed
            20.10.1997 multithread version finished
            31.10.1997 minor improvements
            21.12.1997 two process version programming started
            03.01.1997 first two process version completed
            16.01.1998 constant MAXHEAP defined as HMS_MAXCORN
            26.01.1998 calls to AllocConsole and FreeConsole added
            12.06.1998 minor improvements
            16.06.1998 menu item `Reload Maze...' added
----------------------------------------------------------------------*/
#define _POSIX_SOURCE
#define select SELECT           /* just to fool Window's includes */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <math.h>
#include <time.h>
#include <windows.h>
#undef select
#include "maze.h"
#include "server.h"
#include "sprite.h"
#include "whamster.rh"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define PRGNAME           "whamster"
#define NONAME            "<noname>"

/* --- abbreviations --- */
#define HINST             HINSTANCE
#define MKIRSC            MAKEINTRESOURCE
#define WINSTYLE          WS_TILED|WS_CAPTION|WS_SYSMENU|WS_BORDER
#define WHITE             RGB(0xff, 0xff, 0xff)
#define BLACK             RGB(0x00, 0x00, 0x00)
#define PATH_MAX          _MAX_PATH

/* --- edit box functions --- */
#define edt_set(h,i,s)    SetDlgItemText(h, i, s)
#define edt_get(h,i,s,n)  GetDlgItemText(h, i, s, n)
#define int_set(h,i,n)    SetDlgItemInt(h, i, n, 0)
#define int_get(h,i)      GetDlgItemInt(h, i, NULL, 0)

/* --- check box functions --- */
#define chk_set(h,i,s)    CheckDlgButton(h, i, s)
#define chk_get(h,i)      IsDlgButtonChecked(h, i)

/* --- extensions and values --- */
#define HMS_MINXEXT    8        /* minimal x-extension of maze */
#define HMS_MINYEXT    1        /* minimal y-extension of maze */
#define MAXHEAP     HMS_MAXCORN /* maximal size of corn heap */
#define MINSPEED      0.1F      /* minimal hamster speed */
#define MAXSPEED      100.0F    /* maximal hamster speed */
#define FIELDWD       16        /* width of displayed field (pixels) */
#define SPRITEWD      13        /* width of hamster sprite (pixels) */
#define BDPENALTY   1000        /* penalty for program breakdown */

/* --- flags --- */
#define MZ_LOAD        0        /* load a maze */
#define MZ_SAVE        1        /* save a maze */
#define MZ_NEW         2        /* get a new file name */

/* --- error codes --- */
#define OK             0        /* no error */
#define E_NONE         0        /* no error */
#define E_NOMEM       -1        /* not enough memory */
#define E_FOPEN       -2        /* cannot open file */
#define E_FREAD       -3        /* read error on file */
#define E_FWRITE      -4        /* write error on file */
#define E_POPEN       -5        /* cannot open pipe */
#define E_PREAD       -6        /* read error on pipe */
#define E_PWRITE      -7        /* write error on pipe */
#define E_PBROKEN     -8        /* broken pipe */
#define E_PROCESS     -9        /* cannot start process */
#define E_THREAD     -10        /* cannot start thread */
#define E_UNKNOWN    -11        /* unknown error */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- application specific data */
  int     xext, yext;           /* maze extensions */
  float   wallprob;             /* wall probability */
  int     total;                /* total corn in maze */
  int     maxheap;              /* maximal size of corn heap */
  float   speed;                /* hamster speed (fields/second) */
  int     corncnt;              /* collected corn */
  int     movecnt;              /* number of moves made */
  int     crshcnt;              /* number of crashes against walls */
  int     score;                /* score achieved */
  char    fn_maze[PATH_MAX +256];  /* file name of current maze */
  char    fn_hms [PATH_MAX +256];  /* file name of executable */
} AppData;                      /* (application specific data) */

typedef struct {                /* --- hamster control block --- */
  HAMSTER *hamster;             /* hamster to control */
  SPRITE  *sprite;              /* sprite associated with hamster */
} HMSCB;                        /* (hamster control block) */

/*----------------------------------------------------------------------
  Constants
----------------------------------------------------------------------*/
const char *errmsgs[] = {       /* error messages */
  /* E_NONE      0 */  "no error",
  /* E_NOMEM    -1 */  "not enough memory",
  /* E_FOPEN    -2 */  "cannot open file\n%s",
  /* E_FREAD    -3 */  "read error on file\n%s",
  /* E_FWRITE   -4 */  "write error on file\n%s",
  /* E_POPEN    -5 */  "cannot open pipe",
  /* E_PREAD    -6 */  "read error on pipe",
  /* E_PWRITE   -7 */  "write error on pipe",
  /* E_PBROKEN  -8 */  "broken pipe",
  /* E_PROCESS  -9 */  "cannot start child process\n%s",
  /* E_THREAD  -10 */  "cannot start pipe thread",
  /* E_UNKNOWN -11 */  "unknown error"
};

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
static AppData appdata = {      /* application specific data */
  20, 15,                       /* initial maze extensions */
  0.75F, 500, 8,                /* wall prob., total corn, max. heap */
  2.5F,                         /* inital hamster speed */
  0, 0, 0, 0,                   /* score variables */
  NONAME,                       /* file name of current maze */
  "chamster.exe" };             /* file name of executable */
static HINST   hinst;           /* instance handle */
static HWND    hmain;           /* handle of main window */
static char    buf[PATH_MAX +1024];    /* message buffer */

/* --- brushes --- */
static HBRUSH  br_white;        /* white  brush (to draw background) */
static HBRUSH  br_grey;         /* grey   brush (to draw home field) */
static HBRUSH  br_yellow;       /* yellow brush (to draw corn) */

/* --- pens --- */
static HPEN    pen_white;       /* white  pen (to draw background) */
static HPEN    pen_black;       /* black  pen (to draw walls) */
static HPEN    pen_orange;      /* orange pen (to draw corn) */

/* --- child process and pipe variables --- */
static HANDLE  process = NULL;  /* handle of hamster process */
static HANDLE  thread  = NULL;  /* handle of pipe input thread */
static HANDLE  p_in    = NULL;  /* input  pipe (child  -> parent) */
static HANDLE  p_out   = NULL;  /* output pipe (parent -> child) */

/* --- hamster variables --- */
static HBITMAP hms_bmap;        /* handle of sprite bitmap */
static HBITMAP hms_mask;        /* handle of mask bitmap */
static MAZE    *maze;           /* maze the hamster is in */
static HMSCB   hmscb[1] = {     /* hamster control blocks */
  { NULL, NULL } };             /* (at most one hamster) */
static int     xoff, yoff;      /* offsets for drawing */
                                
/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void error (int code, ...)
{                               /* --- print error message */
  va_list     args;             /* list of variable arguments */
  const char  *msg;             /* error message */

  if ((code > 0) || (code < E_UNKNOWN))
    code = E_UNKNOWN;           /* check error code */
  va_start(args, code);         /* get variable arguments */
  msg = errmsgs[-code];         /* and error message text */
  if (!msg) msg = errmsgs[-E_UNKNOWN];
  vsprintf(buf, msg, args);     /* format error message */
  MessageBox(hmain, buf, PRGNAME, MB_OK|MB_ICONSTOP);
  va_end(args);                 /* show alert dialog box */
}  /* error() */

/*--------------------------------------------------------------------*/

static double drand (void)
{                               /* --- compute random number in [0,1) */
  return rand() /(RAND_MAX +1.0);
}  /* drand() */

/*----------------------------------------------------------------------
  Drawing Functions
----------------------------------------------------------------------*/

static void draw_field (HDC hdc, int x, int y)
{                               /* --- draw maze field */
  RECT  rc;                     /* field rectangle */
  short c, t;                   /* contents of maze field and buffer */
  short cw, ce, cs, cn;         /* contents of surrounding fields */

  c  = mz_getfld(maze, x,   y); /* get contents of */
  cw = mz_getfld(maze, x-1, y); /* field to draw */
  ce = mz_getfld(maze, x+1, y); /* and contents of */
  cs = mz_getfld(maze, x, y-1); /* surrounding fields */
  cn = mz_getfld(maze, x, y+1);
  x  = xoff +x *FIELDWD;        /* multiply coordinates with */
  y  = yoff -y *FIELDWD;        /* field width and add offsets */

  /* --- draw field interior --- */
  rc.left   = x; rc.right = x +FIELDWD -1;
  rc.bottom = y; rc.top   = y -FIELDWD +1;
  FillRect(hdc, &rc, (c & MZ_HOME) ? br_grey : br_white);

  /* --- draw corners --- */
  t = ((c & (MZ_WEST|MZ_SOUTH)) | (cs & MZ_WEST) | (cw & MZ_SOUTH));
  SetPixel(hdc, x -1,         y,          (t) ? BLACK : WHITE);
  t = ((c & (MZ_EAST|MZ_SOUTH)) | (cs & MZ_EAST) | (ce & MZ_SOUTH));
  SetPixel(hdc, x +FIELDWD-1, y,          (t) ? BLACK : WHITE);
  t = ((c & (MZ_EAST|MZ_NORTH)) | (cn & MZ_EAST) | (ce & MZ_NORTH));
  SetPixel(hdc, x +FIELDWD-1, y -FIELDWD, (t) ? BLACK : WHITE);
  t = ((c & (MZ_WEST|MZ_NORTH)) | (cn & MZ_WEST) | (cw & MZ_NORTH));
  SetPixel(hdc, x -1,         y -FIELDWD, (t) ? BLACK : WHITE);

  /* --- draw walls --- */
  SelectObject(hdc, (c & MZ_SOUTH) ? pen_black : pen_white);
  MoveToEx(hdc, x,            y,          NULL);
  LineTo  (hdc, x +FIELDWD-1, y);
  SelectObject(hdc, (c & MZ_EAST)  ? pen_black : pen_white);
  MoveToEx(hdc, x +FIELDWD-1, y -1,       NULL);
  LineTo  (hdc, x +FIELDWD-1, y -FIELDWD);
  SelectObject(hdc, (c & MZ_NORTH) ? pen_black : pen_white);
  MoveToEx(hdc, x,            y -FIELDWD, NULL);
  LineTo  (hdc, x +FIELDWD-1, y -FIELDWD);
  SelectObject(hdc, (c & MZ_WEST)  ? pen_black : pen_white);
  MoveToEx(hdc, x -1,         y -1,       NULL);
  LineTo  (hdc, x -1,         y -FIELDWD);

  /* --- draw corn heap --- */
  c &= MZ_ITEMS;                /* get size of corn heap */
  if (c > 0) {                  /* if there is corn on the field */
    if (c > 12) c = 12;         /* limit heap size */
    t = (c >> 1) +(c & 1);      /* calculate ellipse */
    c = (c >> 1) +1;            /* extensions */
    SelectObject(hdc, pen_orange);  /* get orange pen */
    SelectObject(hdc, br_yellow);   /* and yellow brush */
    Ellipse(hdc, x +(FIELDWD/2)-1 -t, y -(FIELDWD/2) -t,
                 x +(FIELDWD/2)-1 +c, y -(FIELDWD/2) +c);
  }                             /* draw corn heap */
}  /* draw_field() */

/*--------------------------------------------------------------------*/

static void hms_disp (HAMSTER *hms, int cmd, int arg)
{                               /* --- hamster display function */
  HDC    hdc;                   /* handle of device context */
  int    id;                    /* hamster identifier */
  int    x, y, sx, sy;          /* coordinates of hamster */
  int    dir;                   /* direction of hamster */
  SPRITE *sprite;               /* sprite to control */
  int    cnt;                   /* number of steps to move */
  int    series;                /* sprite image series */

  if (appdata.speed >= MAXSPEED)
    return;                     /* if maximal speed set, abort */
  id     = hms_id(hms);         /* get hamster identifier */
  sprite = hmscb[id].sprite;    /* and associated sprite */
  switch (cmd) {                /* evaluate command */
    case HMS_CREATED:           /* --- if hamster was created */
      mz_getpos(maze, &sx, &sy);/* get start position and */
      hms_pos(hms, &x, &y);     /* current position of hamster */
      x += sx; y += sy;         /* and compute absolute position */
      x = xoff +x *FIELDWD +(FIELDWD-SPRITEWD)/2;
      y = yoff -y *FIELDWD -(FIELDWD-SPRITEWD)/2 -SPRITEWD;
      dir = hms_dir(hms);       /* get direction (line of sight) */
      switch (dir) {            /* evaluate direction */
        case HMS_EAST : series = 0; break;
        case HMS_NORTH: series = 1; break;
        case HMS_WEST : series = 2; break;
        case HMS_SOUTH: series = 3; break;
        default       : series = 0; break;
      }                         /* get sprite image series number */
      spr_select(sprite, series, 0);
      spr_draw(sprite, x, y);   /* select and draw sprite picture */

    case HMS_SCORED:            /* --- if score has changed */
      sprintf(buf, "%s %10ld", PRGNAME, hms_score(hms));
      SetWindowText(hmain, buf);
      break;                    /* set new window title */

    case HMS_MOVED:             /* --- if hamster was moved */
      dir = hms_dir(hms);       /* get direction (line of sight) */
      switch (dir) {            /* and evaluate it */
        case HMS_EAST : x =  2; y =  0; cnt = FIELDWD/2; break;
        case HMS_NORTH: x =  0; y = -2; cnt = FIELDWD/2; break;
        case HMS_WEST : x = -2; y =  0; cnt = FIELDWD/2; break;
        case HMS_SOUTH: x =  0; y =  2; cnt = FIELDWD/2; break;
        default       : x =  0; y =  0; cnt = 0;         break;
      }                         /* get offsets to new field */
      if      (appdata.speed >= 16) { x *= 8; y *= 8; cnt /= 8; }
      else if (appdata.speed >=  8) { x *= 4; y *= 4; cnt /= 4; }
      else if (appdata.speed >=  4) { x *= 2; y *= 2; cnt /= 2; }
      spr_move(sprite, x, y, 0, cnt, 1.0F/cnt/appdata.speed);
      break;                    /* redraw and move sprite */

    case HMS_TURNED:            /* --- if hamster was turned */
      dir = hms_dir(hms);       /* get direction (line of sight) */
      switch (dir) {            /* and evaluate it */
        case HMS_EAST : series = (arg == HMS_POS) ? 3 : 5; break;
        case HMS_NORTH: series = (arg == HMS_POS) ? 0 : 6; break;
        case HMS_WEST : series = (arg == HMS_POS) ? 1 : 7; break;
        case HMS_SOUTH: series = (arg == HMS_POS) ? 2 : 4; break;
        default       : series = (arg == HMS_POS) ? 3 : 5; break;
      }                         /* get sprite image series number */
      if (appdata.speed >= 50) {/* if not to show turns */
        spr_select(sprite, series, 6);
        return;                 /* simply select the last image */
      }                         /* and abort the function */
      spr_select(sprite, series, 0);   /* select image series */
      if      (appdata.speed >= 16) { x = 6; cnt = 1; }
      else if (appdata.speed >=  4) { x = 3; cnt = 2; }
      else                          { x = 1; cnt = 6; }
      spr_move(sprite, 0, 0, x, cnt, 1.0F/cnt/appdata.speed);
      break;                    /* redraw and turn sprite */

    case HMS_FIELD:             /* --- if field contents has changed */
      mz_getpos(maze, &sx, &sy);/* get start position and */
      hms_pos(hms, &x, &y);     /* current position of hamster */
      x += sx; y += sy;         /* and compute absolute position */
      spr_undraw(sprite);       /* undraw sprite */
      hdc = GetDC(hmain);       /* get device context for window */
      draw_field(hdc, x, y);    /* and update field */
      ReleaseDC(hmain, hdc);    /* release device context for window */
      spr_draw(sprite, INT_MIN, INT_MIN);
      break;                    /* redraw sprite */

    case HMS_DELETED:           /* --- if hamster is deleted */
      spr_undraw(sprite); break;/* undraw sprite */

    default: break;             /* ignore everything else */
  }  
}  /* hms_disp() */

/*--------------------------------------------------------------------*/

static void resize (HWND hwnd)
{                               /* --- adapt window size */
  RECT rc;                      /* client area rectangle */
  int  x, y, dx, dy;            /* coordinates and offsets */
  
  GetClientRect(hwnd, &rc);     /* get client rectangle of window */
  xoff = 1; yoff = appdata.yext *FIELDWD;  /* and compute offsets */
  rc.right  = rc.left +(dx = appdata.xext *FIELDWD +1);
  rc.bottom = rc.top  +(dy = appdata.yext *FIELDWD +1);
  AdjustWindowRect(&rc, WINSTYLE, 1);
  x = rc.right  -rc.left;       /* get window client area, */
  y = rc.bottom -rc.top;        /* compute new window rectangle, */
  GetWindowRect(hwnd, &rc);     /* and resize window */
  MoveWindow(hwnd, rc.left, rc.top, x, y, 1);
  GetClientRect(hwnd, &rc);     /* get resulting rectangle */
  if ((rc.right  -rc.left != dx)     /* if resulting client area */
  ||  (rc.bottom -rc.top  != dy)) {  /* has a different size */
    x += dx -(rc.right  -rc.left);   /* (e.g. because of a */
    y += dy -(rc.bottom -rc.top);    /* two-line menu) */
    MoveWindow(hwnd, rc.left, rc.top, x, y, 1);
  }                             /* resize window a second time */
  InvalidateRect(hwnd, NULL,0); /* trigger window redraw */
}  /* resize() */

/*----------------------------------------------------------------------
  Dialog Box Functions
----------------------------------------------------------------------*/

BOOL CALLBACK dp_about (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'about' dialog box procedure */
  switch (msg) {                /* evaluate message */
    case WM_COMMAND:            /* --- if command */
      if (wp != IDOK) break;    /* process only 'Ok' button */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, 0);       /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_about() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_size (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'set maze size' dialog box */
  MAZE *tmp;                    /* temporary buffer */
  int  xext, yext;              /* x- and y-extension */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* --- if dialog initialization */
      int_set(hwnd, DI_XEXT, appdata.xext);  /* set edit */
      int_set(hwnd, DI_YEXT, appdata.yext);  /* controls */
      return 1;                 /* add items to list box */
    case WM_COMMAND:            /* --- if command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          xext = int_get(hwnd, DI_XEXT);  /* get input from */
          yext = int_get(hwnd, DI_YEXT);  /* edit controls */
          if (xext < HMS_MINXEXT) xext = HMS_MINXEXT;
          if (xext > HMS_MAXXEXT) xext = HMS_MAXXEXT;
          if (yext < HMS_MINYEXT) yext = HMS_MINYEXT;
          if (yext > HMS_MAXYEXT) yext = HMS_MAXYEXT;
          tmp = mz_create(xext, yext);
          if (!tmp) { error(E_NOMEM); break; }
          mz_delete(maze); maze = tmp;
          appdata.xext = xext;  /* create and set a new maze and */
          appdata.yext = yext;  /* note the new maze extensions */
          EndDialog(hwnd, 0);   /* terminate the dialog */
          return 1;             /* and return 'ok' */
        case IDCANCEL: break;   /* if 'Cancel' button pressed */
        default      : return 0;
      }                         /* everything else not handled */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, -1);      /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_size() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_random (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'randomize maze' dialog box */
  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* --- if dialog initialization */
      int_set(hwnd, DI_WALLPROB, (int)(appdata.wallprob *100 +0.5F));
      int_set(hwnd, DI_TOTAL,    appdata.total);
      int_set(hwnd, DI_MAXHEAP,  appdata.maxheap);
      return 1;                 /* set edit controls */
    case WM_COMMAND:            /* --- if command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          appdata.wallprob = int_get(hwnd, DI_WALLPROB) /100.0F;
          appdata.total    = int_get(hwnd, DI_TOTAL);
          appdata.maxheap  = int_get(hwnd, DI_MAXHEAP);
          if (appdata.wallprob < 0.0F)    appdata.wallprob = 0.0F;
          if (appdata.wallprob > 1.0F)    appdata.wallprob = 1.0F;
          if (appdata.maxheap  < 1)       appdata.maxheap  = 1;
          if (appdata.maxheap  > MAXHEAP) appdata.maxheap  = MAXHEAP;
          mz_init(maze, drand,   appdata.wallprob,
                  appdata.total, appdata.maxheap);
          EndDialog(hwnd, 0);   /* get inputs from edit controls */
          return 1;             /* and reinitialize maze */
        case IDCANCEL: break;   /* if 'Cancel' button pressed */
        default      : return 0;
      }                         /* everything else not handled */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, -1);      /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_random() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_speed (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'set maze size' dialog box */
  static char buf[16];          /* text buffer */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* --- if dialog initialization */
      sprintf(buf, "%g", appdata.speed);
      edt_set(hwnd, DI_SPEED, buf);
      return 1;                 /* add items to dialog box */
    case WM_COMMAND:            /* --- if command */
      switch (wp) {             /* evaluate command code */
        case IDOK:              /* if 'Ok' button pressed */
          edt_get(hwnd, DI_SPEED, buf, 16);
          appdata.speed = (float)strtod(buf, NULL);
          if (appdata.speed < MINSPEED) appdata.speed = MINSPEED;
          if (appdata.speed > MAXSPEED) appdata.speed = MAXSPEED;
          EndDialog(hwnd, 0);   /* get input from edit controls */
          return 1;
        case IDCANCEL: break;   /* if 'Cancel' button pressed */
        default      : return 0;
      }                         /* everything else not handled */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, -1);      /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_speed() */

/*--------------------------------------------------------------------*/

BOOL CALLBACK dp_result (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- 'set maze size' dialog box */
  static char buf[16];          /* text buffer */

  switch (msg) {                /* evaluate message */
    case WM_INITDIALOG:         /* --- if dialog initialization */
      int_set(hwnd, DI_CORNCNT, appdata.corncnt);
      int_set(hwnd, DI_MOVECNT, appdata.movecnt);
      int_set(hwnd, DI_CRSHCNT, appdata.crshcnt);
      sprintf(buf, "%ld", appdata.score);
      edt_set(hwnd, DI_SCORE,   buf);
      return 1;                 /* add items to dialog box */
    case WM_COMMAND:            /* --- if command */
      if (wp != IDOK) break;    /* process only 'Ok' button */
    case WM_CLOSE:              /* --- if dialog closed */
      EndDialog(hwnd, 0);       /* terminate dialog */
      return 1;                 /* return 'ok' */
  }
  return 0;                     /* everything else not handled */
}  /* dp_result() */

/*--------------------------------------------------------------------*/

static int mz_file (int mode)
{                               /* --- execute maze file operations */
  static char fname[_MAX_PATH+1];    /* file name */
  static char path [_MAX_PATH+1];    /* and path */
  static char filter[] =        /* file selection filter */
    "maze (*.maz)\0*.maz\0"     /* mazes */
    "all files (*.*)\0*.*\0";   /* all files */
  static OPENFILENAME ofn = {   /* file selector structure */
    sizeof(OPENFILENAME),       /* lStructSize */
    NULL,   NULL,               /* hwndOwner, hInstance */
    filter, NULL,               /* lpstrFilter, lpstrCustomFilter */
    0, 1,                       /* nMaxCustFilter, nFilterIndex */
    fname, _MAX_PATH,           /* lpstrFile, nMaxFile */
    NULL, _MAX_FNAME+_MAX_EXT,  /* lpstrFileTitle, nMaxFileTitle */
    path,                       /* lpstrInitialDir */
    "Select Maze File",         /* lpstrTitle */
    0, 0, 0,                    /* Flags, nFileOffset, nFileExtension */
    "maz", 0L,                  /* lpstrDefExt, lCustData */
    NULL, NULL };               /* lpfnHook, lpTemplateName */
  FILE *file;                   /* file to load from/save to */
  MAZE *tmp;                    /* temporary buffer */

  if (!(mode & MZ_NEW)          /* if to use old file name */
  &&  (strcmp(appdata.fn_maze, NONAME) != 0))
    strcpy(fname, appdata.fn_maze);
  else {                        /* if to get a new file name */
    ofn.hwndOwner = hmain;      /* set window handle and flags */
    ofn.Flags     = (mode & MZ_SAVE)
      ? OFN_PATHMUSTEXIST : OFN_FILEMUSTEXIST;
    if (GetOpenFileName(&ofn) == 0) 
      return -1;                /* open file selector box */
  }                             /* and get maze file name */
  if (mode & MZ_SAVE) {         /* if to save a maze */
    file = fopen(fname, "w");   /* open selected file */
    if (!file) { error(E_FOPEN, fname); return -1; }
    if (mz_save(maze, file) != 0) error(E_FWRITE, fname);
    fclose(file);               /* save maze and close file */
    strcpy(appdata.fn_maze, fname); }  /* note name of maze */
  else {                        /* if to load a maze */
    file = fopen(fname, "r");   /* open selected file */
    if (!file) { error(E_FOPEN, fname); return -1; }
    tmp = mz_load(file);        /* try to load maze */
    fclose(file);               /* and close input file */
    if (!tmp) { error(E_FREAD, fname); return -1; }
    mz_delete(maze); maze = tmp;/* delete old maze */
    mz_exts(maze, &appdata.xext, &appdata.yext);
    strcpy(appdata.fn_maze, fname);
  }                             /* note size and name of maze */
  return 0;                     /* return 'ok' */
}  /* mz_file() */

/*--------------------------------------------------------------------*/

static int select (void)
{                               /* --- select a hamster program */
  static char fname[_MAX_PATH+1];    /* file name */
  static char path [_MAX_PATH+1];    /* and path */
  static char filter[] =        /* file selection filter */
    "hamster (*.exe)\0*.exe\0"  /* hamsters */
    "all files (*.*)\0*.*\0";   /* all files */
  static OPENFILENAME ofn = {   /* file selector structure */
    sizeof(OPENFILENAME),       /* lStructSize */
    NULL,   NULL,               /* hwndOwner, hInstance */
    filter, NULL,               /* lpstrFilter, lpstrCustomFilter */
    0, 1,                       /* nMaxCustFilter, nFilterIndex */
    fname, _MAX_PATH,           /* lpstrFile, nMaxFile */
    NULL, _MAX_FNAME+_MAX_EXT,  /* lpstrFileTitle, nMaxFileTitle */
    path,                       /* lpstrInitialDir */
    "Select Hamster Program",   /* lpstrTitle */
    OFN_FILEMUSTEXIST, 0, 0,    /* Flags, nFileOffset, nFileExtension */
    "exe", 0L,                  /* lpstrDefExt, lCustData */
    NULL, NULL };               /* lpfnHook, lpTemplateName */

  ofn.hwndOwner = hmain;        /* set window handle */
  if (GetOpenFileName(&ofn) == 0) 
    return -1;                  /* open file selector box */
  strcpy(appdata.fn_hms, fname);/* copy name of hamster program */
  return 0;                     /* return 'ok' */
}  /* select() */

/*----------------------------------------------------------------------
  Process and Thread Functions
----------------------------------------------------------------------*/

static void mn_enable (int enable)
{                               /* --- enable/disable menu items */
  HMENU        menu;            /* handle of window menu */
  MENUITEMINFO mii;             /* for menu customization */
  UINT         flag;            /* enable/disable flag */

  menu = GetMenu(hmain);        /* get menu handle */
  flag = (enable) ? MF_ENABLED : MF_GRAYED;
  EnableMenuItem(menu, MI_RELOAD, flag);
  EnableMenuItem(menu, MI_LOAD,   flag);  /* enable/disable those */
  EnableMenuItem(menu, MI_SAVE,   flag);  /* menu items that should */
  EnableMenuItem(menu, MI_SAVEAS, flag);  /* not be selected while */
  EnableMenuItem(menu, MI_SELECT, flag);  /* a hamster is running */
  EnableMenuItem(menu, MI_SIZE,   flag);
  EnableMenuItem(menu, MI_CLEAR,  flag);
  EnableMenuItem(menu, MI_RANDOM, flag);
  mii.cbSize     = sizeof(MENUITEMINFO);  /* set either the */
  mii.fMask      = MIIM_ID|MIIM_TYPE;     /* 'Start Hamster...' or */
  mii.fType      = MFT_STRING;            /* 'Stop Hamster...' item */
  mii.wID        =  (enable) ? MI_START : MI_STOP;
  LoadString(hinst, (enable) ? CS_START : CS_STOP, buf, 80);
  mii.dwTypeData = buf; mii.cch = strlen(buf);
  SetMenuItemInfo(menu, (enable) ? MI_STOP : MI_START, FALSE, &mii);
}  /* mn_enable() */

/*--------------------------------------------------------------------*/

static void stop (int inthread, int errcode)
{                               /* --- stop hamster thread */
  HAMSTER *hamster;             /* the hamster and */
  SPRITE  *sprite;              /* its associated sprite */

  mn_enable(1);                 /* enable/change menu items */

  /* --- stop process and thread --- */
  if (p_in)  { CloseHandle(p_in);  p_in  = NULL; }
  if (p_out) { CloseHandle(p_out); p_out = NULL; }
  if (process) {                /* if the hamster is running, */
    TerminateProcess(process, 0);    /* terminate the process */
    if (!inthread) {            /* if called from main program */
      TerminateThread(thread, 0); thread = NULL; }
  }                             /* terminate thread and delete it */
  if (errcode != E_NONE)        /* show alert dialog box */
    error(errcode, appdata.fn_hms);
  if (!process) return;         /* if there is no process, abort */
  process = NULL;               /* delete the hamster process */

  /* --- clean up and show result --- */
  hamster = hmscb[0].hamster;   /* get hamster */
  if (!hamster) {               /* if no hamster exists, clear score */
    appdata.corncnt = appdata.movecnt = 0;
    appdata.crshcnt = appdata.score   = 0; }
  else {                        /* if a hamster exists */
    appdata.corncnt = hms_corncnt(hamster);
    appdata.movecnt = hms_movecnt(hamster);
    appdata.crshcnt = hms_crshcnt(hamster);
    appdata.score   = hms_score(hamster);
    hms_delete(hamster); hmscb[0].hamster = NULL;
  }                             /* note score and delete hamster */
  sprite = hmscb[0].sprite;     /* get hamster sprite */
  if (sprite) {                 /* if a sprite exists, */
    spr_undraw(sprite);         /* undraw sprite */
    spr_delete(sprite, 0); hmscb[0].sprite = NULL;
  }                             /* delete sprite */
  DialogBox(hinst, MKIRSC(DB_RESULT), hmain, dp_result);
  if (appdata.speed >= MAXSPEED)     /* if run without display, */
    InvalidateRect(hmain, NULL, 0);  /* redraw window */
  SetWindowText(hmain, PRGNAME);     /* reset window title */
  if (inthread) ExitThread(0);  /* terminate thread */
}  /* stop() */

/*--------------------------------------------------------------------*/

static int readarg (int *arg)
{                               /* --- read argument from pipe */
  char  c;                      /* character read */
  DWORD cnt;                    /* number of bytes read */
  int   sign  = 0;              /* sign of the argument */
  int   state = 0;              /* state of automaton */

  *arg = 0;                     /* initialize argument */
  while (1) {                   /* read next character */
	if (!ReadFile(p_in, &c, 1, &cnt, NULL)) stop(1, E_PREAD);
    if (cnt <= 0)                           stop(1, E_PBROKEN);
    switch (state) {            /* evaluate state of automaton */
      case 0:                   /* --- skip blanks */
        if      ((c == '\r') || (c == '\n'))   return -1;
        else if (c == '+')    {                state = 1; }
        else if (c == '-')    { sign = -1;     state = 1; }
        else if (isdigit(c))  { *arg = c -'0'; state = 2; }
        else if (!isspace(c)) stop(1, E_PREAD);
        break;                  /* evaluate the character read */
      case 1:                   /* --- sign (+/-) read */
        if (!isdigit(c)) stop(1, E_PREAD);
        *arg = c -'0'; state = 2;  /* check for a digit and */
        break;                     /* continue with state 2 */
      case 2:                   /* --- at least one digit read */
        if      (isdigit(c))  *arg = *arg *10 +(c -'0');
        else if (!isspace(c)) stop(1, E_PREAD);
        else {                  /* if end of argument reached */
          if (sign < 0) *arg = -*arg;
          return ((c == '\r') || (c == '\n')) ? 1 : 0;
        } break;                /* set argument value and */
    }                           /* return end of line indicator */
  }  /* while (1) .. */
}  /* readarg() */

/*----------------------------------------------------------------------
The above function is necessary to read the command arguments from the
pipe. A redirection of the standard input (with SetStdHandle) does not
work, since it does not have any effect on the `scanf' function (for
some unknown reason --- it works fine for the child process). Checking
for a broken pipe by testing the value of `cnt' (as done above) does
not work, too, since the read remains blocked if the connected process
is killed (in contrast to UNIX, where killing a process causes an end
of file state of all pipes connected to it).
----------------------------------------------------------------------*/

static DWORD ip_pipe (LPDWORD dummy)
{                               /* --- pipe input procedure */
  static char buf[96];          /* write buffer */
  HAMSTER *hamster;             /* hamster to control */
  char    cmd;                  /* command read */
  int     id = 0;               /* hamster identifier */
  int     arg;                  /* command argument */
  int     x, y;                 /* hamster position */
  DWORD   cnt;                  /* number of read/written chars */

  while (1) {                   /* pipe input loop */

    /* --- read command from pipe --- */
    do {                        /* read next character */
      if (!ReadFile(p_in, &cmd, 1, &cnt, NULL)) stop(1, E_PREAD);
      if (cnt <= 0)                             stop(1, E_PBROKEN);
    } while (isspace(cmd));     /* until command read */

    /* --- read command argument(s) from pipe --- */
    switch (cmd) {              /* read command arguments */
      case 'c':                 /* --- create hamster */
        if (readarg(&arg) >= 0) stop(1, E_PREAD);
        break;                  /* no arguments */
      case 'd': case 'm':       /* --- delete and move hamster */
        if (readarg(&id)  != 1) stop(1, E_PREAD);
        break;                  /* one argument: hamster id */
      case 't': case 'l':       /* --- turn hamster and load */
        if ((readarg(&id) != 0) || (readarg(&arg) != 1))
          stop(1, E_UNKNOWN);   /* two arguments: hamster id */
        break;                  /* and direction/amount */
      default:                  /* --- unknown command */
        stop(1, E_PREAD); return 0;
    }                           /* stop execution */

    /* --- evaluate command --- */
    hamster = hmscb[id].hamster;/* get hamster to control */
    switch (cmd) {              /* evaluate command read */
      case 'c':                 /* --- create hamster */
        if (hamster) {          /* if a hamster already exists, */
          strcpy(buf, "c -1\n"); break; }       /* deny request */
        hmscb[0].sprite =       /* create a sprite */
          spr_create(hmain, SPRITEWD, SPRITEWD, hms_bmap, hms_mask);
        if (!hmscb[0].sprite) stop(1, E_NOMEM);
        hmscb[0].hamster =      /* create a hamster */
        hamster = hms_create(maze, 0, HMS_EAST, hms_disp);
        if (!hamster) stop(1, E_NOMEM);
        hms_pos(hamster,&x,&y); /* get hamster position */
        sprintf(buf, "c %d %d %d %d %d %d %d\n",
          hms_id(hamster), x, y, hms_dir(hamster),
          hms_look(hamster), hms_corn(hamster), hms_load(hamster));
        break;                  /* send hamster data */

      case 'd':                 /* --- delete hamster */
        strcpy(buf, "d 0\n");   /* echo delete command */
        WriteFile(p_out, buf, strlen(buf), &cnt, NULL);
        FlushFileBuffers(p_out);
        stop(1, OK); return 0;  /* stop hamster process */

      case 'm':                 /* --- move hamster */
        hms_move(hamster);      /* move hamster forward */
        hms_pos(hamster,&x,&y); /* and get new position */
        sprintf(buf, "m %d %d %d %d %d\n",
          hms_id(hamster), x, y, hms_look(hamster), hms_corn(hamster));
        break;                  /* send new hamster data */

      case 't':                 /* --- turn hamster */
        hms_turn(hamster, arg); /* turn hamster around */
        sprintf(buf, "t %d %d %d\n",
          hms_id(hamster), hms_dir(hamster), hms_look(hamster));
        break;                  /* send new hamster data */

      case 'l':                 /* --- take/drop corn */
        hms_take(hamster, arg); /* take/drop some corn */
        sprintf(buf, "l %d %d %d\n",
          hms_id(hamster), hms_corn(hamster), hms_load(hamster));
        break;                  /* send new hamster data */

      default:                  /* --- unknown command */
        stop(1, E_PREAD); return 0;  /* stop execution */
    }
    if (!WriteFile(p_out, buf, strlen(buf), &cnt, NULL))
      stop(1, E_PWRITE);        /* write reply to command to pipe */
  }  /* while (1) .. */
}  /* ip_pipe() */

/*--------------------------------------------------------------------*/

static void start (void)
{                               /* --- start hamster process */
  SECURITY_ATTRIBUTES sa;       /* security attributes */
  PROCESS_INFORMATION pi;       /* process information */
  STARTUPINFO         si;       /* startup information */
  HANDLE     c_in, c_out;       /* child  ends of pipes */
  DWORD               id;       /* thread identifier */
  int                 ok;       /* flag for success */

  if (process) return;          /* if hamster is running, abort */
  mn_enable(0);                 /* disable/change menu items */

  /* --- create pipes --- */
  sa.nLength              = sizeof(SECURITY_ATTRIBUTES);
  sa.bInheritHandle       = TRUE;   /* initialize security */
  sa.lpSecurityDescriptor = NULL;   /* attributes structure */
  if (!CreatePipe(&p_in, &c_out, &sa, 0)) {
    stop(0, E_POPEN); return; } /* create pipe child  -> parent */
  if (!CreatePipe(&c_in, &p_out, &sa, 0)) {
    CloseHandle(c_out);         /* create pipe parent -> child */
    stop(0, E_POPEN); return;   /* and then set the child ends */
  }                             /* of the pipes as standard handles */
  AllocConsole();               /* for a newly created console */
  SetStdHandle(STD_OUTPUT_HANDLE, c_out);
  SetStdHandle(STD_INPUT_HANDLE,  c_in);

  /* --- start hamster process --- */
  si.cb          = sizeof(STARTUPINFO); 
  si.lpReserved  = NULL;        /* initialize  */
  si.lpReserved2 = NULL;        /* process     */
  si.cbReserved2 = 0;           /* startup     */
  si.lpDesktop   = NULL;        /* information */
  si.lpTitle     = appdata.fn_hms;
  si.dwFlags     = 0; 
  ok = CreateProcess(NULL,      /* create a child process */
                     appdata.fn_hms, NULL, NULL, TRUE,
                     0, NULL, NULL, &si, &pi);
  FreeConsole();                /* free console and close handles */
  CloseHandle(c_out); CloseHandle(c_in);
  if (!ok) { stop(0, E_PROCESS); return; }
  process = pi.hProcess;        /* note the process handle */

  /* --- start pipe input thread --- */
  thread = CreateThread(NULL,   /* create pipe input thread */
                        0, (LPTHREAD_START_ROUTINE)ip_pipe,
                        NULL, 0, &id);
  if (!thread) stop(0, E_THREAD);
}  /* start() */

/*----------------------------------------------------------------------
  Window Function
----------------------------------------------------------------------*/

LRESULT CALLBACK wp_main (HWND hwnd, UINT msg, WPARAM wp, LPARAM lp)
{                               /* --- main window procedure */
  static PAINTSTRUCT  ps;       /* for BeginPaint call */
  static MENUITEMINFO mii;      /* for menu customization */
  HDC    hdc;                   /* handle of device context */
  int    x, y, dx, dy;          /* coordinates and distances */
  short  t, chg;                /* temporary buffer */

  switch (msg) {                /* evaluate message */
    case WM_PAINT:              /* --- if to paint window */
      if (process) {            /* if hamster process is running, */
        SuspendThread(thread);  /* suspend thread while drawing */
        if (hmscb[0].sprite)    /* if a hamster sprite exists */
          spr_undraw(hmscb[0].sprite);
      }                         /* undraw sprite (for savety) */
      hdc = BeginPaint(hwnd, &ps);
      if (!hdc) return 0;       /* get device context for drawing */
      mz_exts(maze, &dx, &dy);  /* get maze extensions */
      for (x = dx; --x >= 0; )  /* and traverse fields */
        for (y = dy; --y >= 0; )
          draw_field(hdc, x, y);/* draw maze fields */
      SetWindowText(hmain, PRGNAME);
      EndPaint(hwnd, &ps);      /* release device context */
      if (process) {            /* if hamster process is running, */
        if (hmscb[0].sprite)    /* if a hamster sprite exists */
          spr_draw(hmscb[0].sprite, INT_MIN, INT_MIN);
        ResumeThread(thread);   /* redraw hamster sprite and */
      } break;                  /* resume suspended thread */

    case WM_INITMENUPOPUP:      /* --- if menu opened */
      SetWindowText(hmain, PRGNAME); return 0;

    case WM_COMMAND:            /* --- if menu item selected */
      switch (wp) {             /* evaluate item */
        case MI_RELOAD:         /* Menu Item `Reload Maze...' */
          if (mz_file(MZ_LOAD)        == 0) resize(hwnd);
          return 0;             /* resize window */

        case MI_LOAD:           /* Menu Item `Load Maze...' */
          if (mz_file(MZ_LOAD|MZ_NEW) == 0) resize(hwnd);
          return 0;             /* resize window */
        
        case MI_SAVE:           /* Menu Item `Save Maze...' */
          mz_file(MZ_SAVE);     /* save maze under old name */
          return 0;

        case MI_SAVEAS:         /* Menu Item `Save Maze As...' */
          mz_file(MZ_SAVE|MZ_NEW);
          return 0;             /* select file and save maze */
        
        case MI_SELECT:         /* Menu Item `Select Hamster...' */
          select(); return 0;   /* select a hamster program */

        case MI_QUIT:           /* Menu Item 'Quit' */
          if (process) {        /* if a hamster is running */
            TerminateProcess(process, 0); process = NULL; }
          DestroyWindow(hwnd);  /* stop hamster process */
          return 0;             /* and destroy window */
        
        case MI_REDRAW:         /* Menu Item 'Redraw' */
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* invalidate window contents */
        
        case MI_SIZE:           /* Menu Item 'Set Maze Size...' */
          if (DialogBox(hinst, MKIRSC(DB_SIZE), hwnd, dp_size) == 0)
            resize(hwnd);       /* if a new maze size was set, */
          return 0;             /* resize window */
        
        case MI_RANDOM:         /* Menu Item 'Randomize Maze...' */
          if (DialogBox(hinst, MKIRSC(DB_RANDOM), hwnd, dp_random) == 0)
            InvalidateRect(hwnd, NULL, 0);
          return 0;             /* initialize and redraw maze */
        
        case MI_CLEAR:          /* Menu Item 'Clear Maze' */
          mz_init(maze, 0, 0.0F, 0, 0);
          InvalidateRect(hwnd, NULL, 0);
          return 0;             /* clear and redraw maze */
        
        case MI_SPEED:          /* Menu Item 'Set Hamster Speed...' */
          DialogBox(hinst, MKIRSC(DB_SPEED), hwnd, dp_speed);
          return 0;             /* get hamster speed */
        
        case MI_START:          /* Menu Item `Start Hamster' */
          start(); return 0;    /* start hamster process */

        case MI_STOP:           /* Menu Item `Stop Hamster' */
          stop(0,OK); return 0; /* stop hamster process */
        
        case MI_ABOUT:          /* Menu Item 'About' */
          DialogBox(hinst, MKIRSC(DB_ABOUT), hwnd, dp_about);
          return 0;             /* show dialog box 'About WHamster' */
      } break;

    case WM_LBUTTONDOWN:        /* if left mouse button pressed */
      if (thread) break;        /* if hamster is running, abort */
      hdc = GetDC(hwnd);        /* get handle of device context */
      x = LOWORD(lp) -xoff;     /* compute coordinates */
      y = yoff -HIWORD(lp);     /* relative to lower left corner */
      if (wp & MK_CONTROL) {    /* if control key pressed, */
        x /= FIELDWD; y /= FIELDWD;    /* remove all walls */
        mz_setfld(maze, x, y, MZ_WALLS, 0); }
      else if (wp & MK_SHIFT) { /* if shift key pressed */
        mz_getpos(maze, &dx, &dy);
        x /= FIELDWD; y /= FIELDWD;   /* get coordinates of old field, */
        mz_setpos(maze, x, y);        /* compute coordinates of new */
        draw_field(hdc, dx, dy); }    /* and redraw old field */
      else {                    /* if shift key not pressed */
        dx = x & 0x0f; dy = y & 0x0f; t = 0;
        if (dx >= FIELDWD/2) { t |= MZ_EAST;  dx = FIELDWD-1 -dx; }
        else                 { t |= MZ_WEST;  }
        if (dy >= FIELDWD/2) { t |= MZ_NORTH; dy = FIELDWD-1 -dy; }
        else                 { t |= MZ_SOUTH; }
        t &= (dx <= dy) ? MZ_WEST|MZ_EAST : MZ_SOUTH|MZ_NORTH;
        dx = (t & MZ_WEST)  ? -1 : 0;  /* get offsets to */
        dy = (t & MZ_SOUTH) ? -1 : 0;  /* next field */
        x /= FIELDWD; y /= FIELDWD;    /* compute field coordinates */
        t ^= mz_getfld(maze, x, y);
        mz_setfld(maze, x, y, MZ_WALLS, t);
      }                         /* toggle wall flag */
      draw_field(hdc, x, y);    /* redraw changed field */
      ReleaseDC(hwnd, hdc);     /* and clear info field */
      SetWindowText(hmain, PRGNAME); break;
      
    case WM_MBUTTONDOWN:        /* if middle mouse button pressed */
    case WM_RBUTTONDOWN:        /* or if right mouse button pressed */
      if (thread) break;        /* if hamster is running, abort */
      x = LOWORD(lp) -xoff;     /* compute coordinates */
      y = yoff -HIWORD(lp);     /* relative to lower left corner */
      x /= FIELDWD; y /= FIELDWD;
      mz_getpos(maze, &dx,&dy); /* get initial position */
      if ((x == dx) && (y == dy))
        break;                  /* do not place corn on home field */
      if (wp & MK_CONTROL)      /* if control key pressed, */
        t = chg = 0;            /* clear field */
      else {                    /* if control key not pressed */
        t   = mz_getfld(maze, x, y) & MZ_ITEMS;
        chg = (wp & MK_SHIFT) ? 5 : 1;
        if (msg == WM_MBUTTONDOWN) chg = -chg;
      }                         /* get change of corn heap */
      t = (chg == 0) ? 0 : ((t +chg +(MAXHEAP+1)) %(MAXHEAP+1));
      mz_setfld(maze, x, y, MZ_ITEMS, t);
      hdc = GetDC(hwnd);        /* set new corn heap */
      draw_field(hdc, x, y);    /* and redraw field */
      ReleaseDC(hwnd, hdc);     /* release device context */
      sprintf(buf, "%s %10d", PRGNAME, t);
      SetWindowText(hmain, buf);/* show new heap size */
      break;                    /* in window title */

    case WM_CLOSE:              /* window closed */
    case WM_DESTROY:            /* or destroyed */
      stop(0, E_NONE);          /* stop hamster thread */
      PostQuitMessage(0);       /* terminate program */
      return 0;
  }
  return DefWindowProc(hwnd, msg, wp, lp);
}  /* wp_main() */

/*----------------------------------------------------------------------
  Main Program
----------------------------------------------------------------------*/

int WINAPI WinMain (HINSTANCE hcurr, HINSTANCE hprev,
                    LPSTR cmdln, int cmd)
{                               /* --- main function */
  WNDCLASS wc;                  /* window class */
  RECT     rc;                  /* window rectangle */
  MSG      msg;                 /* message */

  /* --- create window class --- */
  hinst = hcurr;                /* note instance handle */
  if (!hprev) {                 /* if first instance of this program */
    wc.style         = CS_HREDRAW | CS_VREDRAW;
    wc.lpfnWndProc   = wp_main;
    wc.cbClsExtra    = wc.cbWndExtra = 0;
    wc.hInstance     = hinst;
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = GetStockObject(WHITE_BRUSH);
    wc.lpszMenuName  = MKIRSC(MN_MAIN);
    wc.lpszClassName = "HamsterClass";
    if (!RegisterClass(&wc))    /* register class */
      return 0;                 /* of main window */
  }

  /* --- create a maze --- */
  srand((unsigned)time(NULL));  /* initialize random number generator */
  maze = mz_create(appdata.xext, appdata.yext);
  if (!maze) { error(E_NOMEM); return 0; }
  xoff = 1; yoff = appdata.yext *FIELDWD;

  /* --- get brushes --- */
  br_white   = GetStockObject(WHITE_BRUSH);
  br_grey    = GetStockObject(LTGRAY_BRUSH);
  br_yellow  = CreateSolidBrush(RGB(0xff, 0xff, 0x00));

  /* --- get pens --- */
  pen_white  = GetStockObject(WHITE_PEN);
  pen_black  = GetStockObject(BLACK_PEN);
  pen_orange = CreatePen(PS_SOLID, 1, RGB(0xff, 0xa5, 0x00));
      
  /* --- load bitmaps --- */
  hms_bmap = LoadBitmap(hinst, MKIRSC(BM_HAMSTER));
  if (!hms_bmap) return 0;      /* load hamster and mask bitmap */
  hms_mask = LoadBitmap(hinst, MKIRSC(BM_MASK));
  if (!hms_mask) { DeleteObject(hms_bmap); return 0; }

  /* --- create window --- */
  rc.left   = rc.top = 0;       /* compute initial window size */
  rc.right  = appdata.xext *FIELDWD +1;
  rc.bottom = appdata.yext *FIELDWD +1;
  AdjustWindowRect(&rc, WINSTYLE, 1);
  hmain = CreateWindow("HamsterClass", PRGNAME, WINSTYLE,
                       30, 30, rc.right-rc.left, rc.bottom-rc.top,
                       NULL, NULL, hinst, NULL);
  if (!hmain) return 0;         /* create main window */
  ShowWindow(hmain, cmd);       /* open and */
  UpdateWindow(hmain);          /* update main window */

  /* --- dispatch messages --- */
  while (GetMessage(&msg, NULL, 0, 0) > 0) {
    TranslateMessage(&msg);     /* get messages, translate */
    DispatchMessage (&msg);     /* and dispatch them */
  }

  /* --- clean up --- */
  DeleteObject(br_yellow);      /* delete created brush */
  DeleteObject(pen_orange);     /* delete created pen */
  DeleteObject(hms_bmap);       /* delete loaded bitmaps */
  DeleteObject(hms_mask);
  mz_delete(maze);              /* delete maze */
  return msg.wParam;
}  /* WinMain() */
